import json

from rest_framework.renderers import (
    INDENT_SEPARATORS,
    LONG_SEPARATORS,
    SHORT_SEPARATORS,
    JSONRenderer,
)

from drf_react_template.schema_form_encoder import SerializerEncoder


class JSONSerializerRenderer(JSONRenderer):
    encoder_class = SerializerEncoder

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render `data` into JSON, returning a bytestring.
            Include renderer_context in JsonEncoder
        """
        if data is None:
            return bytes()

        renderer_context = renderer_context or {}
        indent = self.get_indent(accepted_media_type, renderer_context)

        if indent is None:
            separators = SHORT_SEPARATORS if self.compact else LONG_SEPARATORS
        else:
            separators = INDENT_SEPARATORS
        ret = json.dumps(
            data,
            cls=self.encoder_class,
            indent=indent,
            ensure_ascii=self.ensure_ascii,
            allow_nan=not self.strict,
            separators=separators,
            renderer_context=renderer_context,
        )
        ret = ret.replace('\u2028', '\\u2028').replace('\u2029', '\\u2029')
        return bytes(ret.encode('utf-8'))



import re
from typing import Any, Dict, List, Tuple, Union

from django.conf import settings
from django.core import validators
from django.core.serializers.json import DjangoJSONEncoder
from rest_framework import fields, serializers
from rest_framework import validators as drf_validators

SerializerType = Union[
    serializers.BaseSerializer,
    serializers.Serializer,
    serializers.ListSerializer,
    serializers.Field,
]

SCHEMA_OVERRIDE_KEY = 'schema:override'
UI_SCHEMA_OVERRIDE_KEY = 'uiSchema:override'
COLUMN_PROCESSOR_OVERRIDE_KEY = 'column:override'
OVERRIDE_KEYS = {
    SCHEMA_OVERRIDE_KEY,
    UI_SCHEMA_OVERRIDE_KEY,
    COLUMN_PROCESSOR_OVERRIDE_KEY,
}

DEPENDENCY_SIMPLE_KEY = 'schema:dependencies:simple'
DEPENDENCY_CONDITIONAL_KEY = 'schema:dependencies:conditional'
DEPENDENCY_DYNAMIC_KEY = 'schema:dependencies:dynamic'
DEPENDENCY_OVERRIDE_KEY = 'schema:dependencies:override'
DEPENDENCY_KEYS = {
    DEPENDENCY_SIMPLE_KEY,
    DEPENDENCY_CONDITIONAL_KEY,
    DEPENDENCY_DYNAMIC_KEY,
    DEPENDENCY_OVERRIDE_KEY,
}

STYLE_KEYS_TO_IGNORE = {
    *OVERRIDE_KEYS,
    *DEPENDENCY_KEYS,
}

VALIDATION_MAP = {
    validators.MaxLengthValidator: ['maxLength', lambda v: v.limit_value],
    validators.MinLengthValidator: ['minLength', lambda v: v.limit_value],
    validators.MaxValueValidator: ['maximum', lambda v: v.limit_value],
    validators.MinValueValidator: ['minimum', lambda v: v.limit_value],
    validators.RegexValidator: ['pattern', lambda v: v.regex.pattern],
}

EXCLUDED_VALIDATOR_CLASSES = [
    validators.ProhibitNullCharactersValidator,
    drf_validators.ProhibitSurrogateCharactersValidator,
]


class ProcessingMixin:
    TYPE_MAP: Dict[str, Dict[str, str]] = {
        'CharField': {'type': 'string'},
        'IntegerField': {'type': 'integer', 'widget': 'updown'},
        'FloatField': {'type': 'number', 'widget': 'updown'},
        'DecimalField': {'type': 'number', 'widget': 'updown'},
        'BooleanField': {'type': 'boolean'},
        'DateTimeField': {'type': 'string', 'widget': 'date-time'},
        'DateField': {'type': 'string', 'widget': 'date'},
        'URLField': {'type': 'string', 'widget': 'uri'},
        'ChoiceField': {'type': 'string', 'enum': 'choices'},
        'EmailField': {'type': 'string', 'widget': 'email'},
        'RegexField': {'type': 'string', 'widget': 'regex'},
        'ImageField': {'type': 'file', 'widget': 'file'},
        'ListField': {'type': 'array'},
    }

    def __init__(
        self,
        serializer: SerializerType,
        renderer_context: Dict[str, Any],
        prefix: str = '',
        extra_types: Dict[str, Any] = {},
    ):
        self.serializer = serializer
        if self._is_list_serializer(serializer):
            self.fields = self._filter_fields(serializer.child.fields.items())
        else:
            self.fields = self._filter_fields(serializer.fields.items())
        self.renderer_context = renderer_context
        self.prefix = prefix
        self.extra_types = extra_types
        self.extra_types.update(getattr(settings, 'DRF_REACT_TEMPLATE_TYPE_MAP', {}))
        self.TYPE_MAP.update(self.extra_types)

    def _get_type_map_value(self, field: SerializerType):
        result = {
            'type': field.style.get('schema:type'),
            'enum': field.style.get('schema:enum'),
            'widget': field.style.get('ui:widget'),
        }
        result_default = self.TYPE_MAP.get(type(field).__name__, {})
        for k, v in result_default.items():
            if not result[k]:
                result[k] = result_default[k]
        return result

    def _generate_data_index(self, name: str) -> str:
        return f'{self.prefix}.{name}' if self.prefix else name

    def _get_title(self, field: SerializerType, name: str) -> str:
        result = field.label
        if result is None:
            result = name.title().replace('_', ' ').replace('.', ': ')
        return result

    @staticmethod
    def _filter_fields(all_fields: Tuple) -> Tuple:
        return tuple((name, field) for name, field in all_fields if not field.read_only)

    @staticmethod
    def _is_field_serializer(obj: Any) -> bool:
        return isinstance(obj, serializers.BaseSerializer)

    @staticmethod
    def _is_list_serializer(obj: Any) -> bool:
        return isinstance(obj, serializers.ListSerializer)

    def _is_hidden_serializer(self) -> bool:
        return all(
            [
                self._get_type_map_value(field).get('widget') == 'hidden'
                for name, field in self.fields
            ]
        )


class SchemaProcessor(ProcessingMixin):
    def __init__(
        self,
        serializer: SerializerType,
        renderer_context: Dict[str, Any],
        prefix: str = '',
        extra_types: Dict[str, Any] = {},
    ):
        super().__init__(serializer, renderer_context, prefix, extra_types)
        self.fields_to_be_removed = set()
        self.fields_to_be_kept = set()

    def _is_serializer_optional(self) -> bool:
        return (
            self.serializer.allow_empty
            or (not self.serializer.required)
            or self.serializer.allow_null
        )

    def _get_serializer_title(self) -> str:
        label = self.serializer.label
        if label is not None:
            return label
        if self._is_hidden_serializer():
            return ''
        if self._is_list_serializer(self.serializer):
            class_name = type(self.serializer.child).__name__
        else:
            class_name = type(self.serializer).__name__
        class_name = class_name.replace('Serializer', '')
        class_name = ' '.join(re.findall('[A-Z][^A-Z]*', class_name))
        return class_name

    def _required_fields(self) -> List[str]:
        return [
            name
            for name, field in self.fields
            if field.required and not self._is_field_serializer(field)
        ]

    def _set_validation_properties(
        self, field: SerializerType, result: Dict[str, Any]
    ) -> Dict[str, Any]:
        for validator in field.validators:
            for validator_class, attr_value in VALIDATION_MAP.items():
                if isinstance(validator, validator_class):
                    result_key, result_filter = attr_value
                    result[result_key] = result_filter(validator)

        return result

    def _get_field_properties(self, field: SerializerType, name: str) -> Dict[str, Any]:
        result = {}
        type_map_obj = self._get_type_map_value(field)
        result['type'] = type_map_obj['type']
        result['title'] = self._get_title(field, name)
        if isinstance(field, serializers.ListField):
            if field.allow_empty:
                result['required'] = not getattr(field, 'allow_empty', True)
            result['items'] = self._get_field_properties(field.child, "")
            result['uniqueItems'] = True
        else:
            if field.allow_null:
                result['type'] = [result['type'], 'null']
            enum = type_map_obj.get('enum')
            if enum:
                if enum == 'choices':
                    choices = field.choices
                    result['enum'] = list(choices.keys())
                    result['enumNames'] = [v for v in choices.values()]
                if isinstance(enum, (list, tuple)):
                    if isinstance(enum, (list, tuple)):
                        result['enum'] = [item[0] for item in enum]
                        result['enumNames'] = [item[1] for item in enum]
                    else:
                        result['enum'] = enum
                        result['enumNames'] = [item for item in enum]
            try:
                result['default'] = field.get_default()
            except fields.SkipField:
                pass

        result = self._set_validation_properties(field, result)

        return result

    def _get_all_field_properties(self) -> Dict[str, Any]:
        result = {}
        for name, field in self.fields:
            if self._is_field_serializer(field):
                result[name] = SchemaProcessor(
                    field, self.renderer_context, prefix=self._generate_data_index(name)
                ).get_schema()
            else:
                override = field.style.get(SCHEMA_OVERRIDE_KEY)
                result[name] = override or self._get_field_properties(field, name)
        return result

    def _remove_from_required(
        self, schema: Dict[str, Any], field_name: str
    ) -> Dict[str, Any]:
        try:
            if self._is_list_serializer(self.serializer):
                schema['items']['required'].remove(field_name)
            else:
                schema['required'].remove(field_name)
        except ValueError:
            pass
        return schema

    def _get_from_properties(
        self, schema: Dict[str, Any], field_name: str, pop: bool = False
    ) -> Dict[str, Any]:
        if self._is_list_serializer(self.serializer):
            properties = (
                schema['items']['properties'].pop(field_name)
                if pop
                else schema['items']['properties'].get(field_name)
            )
        else:
            properties = (
                schema['properties'].pop(field_name)
                if pop
                else schema['properties'].get(field_name)
            )
        return properties

    def _simple_dependency(
        self, dependent_properties: Union[str, List[str]], schema: Dict[str, Any]
    ) -> Tuple[List[str], Dict[str, Any]]:
        if not isinstance(dependent_properties, list):
            dependent_properties = [dependent_properties]
        for field_name in dependent_properties:
            schema = self._remove_from_required(schema, field_name)
            self.fields_to_be_kept.add(field_name)
        return dependent_properties, schema

    def _conditional_dependency(
        self, dependent_properties: Union[str, List[str]], schema: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        if not isinstance(dependent_properties, list):
            dependent_properties = [dependent_properties]
        dependency_object = {'properties': {}, 'required': []}
        for field_name in dependent_properties:
            self.fields_to_be_removed.add(field_name)
            schema = self._remove_from_required(schema, field_name)
            properties = self._get_from_properties(schema, field_name, pop=False)
            dependency_object['properties'][field_name] = properties
            dependency_object['required'].append(field_name)
        return dependency_object, schema

    @staticmethod
    def _create_enum_dependency_object(
        field_name: str, enum_key: str, main_properties: Dict[str, Any]
    ) -> Dict[str, Any]:
        if enum_key not in main_properties['enum']:
            raise KeyError(
                f"'{enum_key}' is not a valid enum, the options are: "
                f"{main_properties['enum']}"
            )
        idx = main_properties['enum'].index(enum_key)
        enum_dependency_object = {
            'properties': {field_name: main_properties.copy()},
            'required': [],
        }
        enum_dependency_object['properties'][field_name]['enum'] = [enum_key]
        enum_dependency_object['properties'][field_name]['enumNames'] = [
            main_properties['enumNames'][idx]
        ]

        return enum_dependency_object

    def _dynamic_dependency(
        self, name: str, dependent_properties: Dict[str, Any], schema: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        main_properties = self._get_from_properties(schema, name, pop=False)
        if 'enum' not in main_properties:
            raise KeyError('Only enumerable fields can have dynamic dependencies')
        dependency_object = {'oneOf': []}
        for enum_key, dep_fields in dependent_properties.items():
            enum_dependency_object = self._create_enum_dependency_object(
                name, enum_key, main_properties
            )
            if not dep_fields:
                dep_fields = []
            elif not isinstance(dep_fields, list):
                dep_fields = [dep_fields]
            for field_name in dep_fields:
                self.fields_to_be_removed.add(field_name)
                schema = self._remove_from_required(schema, field_name)
                properties = self._get_from_properties(schema, field_name, pop=False)
                enum_dependency_object['properties'][field_name] = properties
                enum_dependency_object['required'].append(field_name)
            dependency_object['oneOf'].append(enum_dependency_object)
        return dependency_object, schema

    def _add_dependencies(self, schema: Dict[str, Any]) -> Dict[str, Any]:
        dependencies = {}
        for name, field in self.fields:
            dependency_object = {}
            style_keys = set(field.style.keys())
            dep_key = list(DEPENDENCY_KEYS.intersection(style_keys))
            if dep_key:
                if len(dep_key) > 1:
                    raise KeyError(
                        f"Cannot have multiple types of dependencies on a field."
                        f"Please select one of: '{DEPENDENCY_SIMPLE_KEY}', "
                        f"'{DEPENDENCY_CONDITIONAL_KEY}', '{DEPENDENCY_OVERRIDE_KEY}'"
                    )
                dep_key = dep_key[0]
            else:
                continue
            dependent_properties = field.style[dep_key]

            if dep_key == DEPENDENCY_SIMPLE_KEY:
                dependency_object, schema = self._simple_dependency(
                    dependent_properties, schema
                )
            elif dep_key == DEPENDENCY_CONDITIONAL_KEY:
                dependency_object, schema = self._conditional_dependency(
                    dependent_properties, schema
                )
            elif dep_key == DEPENDENCY_DYNAMIC_KEY:
                dependency_object, schema = self._dynamic_dependency(
                    name, dependent_properties, schema
                )
            elif dep_key == DEPENDENCY_OVERRIDE_KEY:
                dependency_object = dependent_properties
            dependencies[name] = dependency_object

        for field_name in self.fields_to_be_removed.difference(self.fields_to_be_kept):
            self._get_from_properties(schema, field_name, pop=True)  # In place mutation
        if dependencies:
            schema['dependencies'] = dependencies
        return schema

    def get_schema(self) -> Dict[str, Any]:
        if self._is_list_serializer(self.serializer):
            schema = {
                'title': self._get_serializer_title(),
                'type': 'array',
                'minItems': 0 if self._is_serializer_optional() else 1,
                'items': {
                    'type': 'object',
                    'required': self._required_fields(),
                    'properties': self._get_all_field_properties(),
                },
            }
        else:
            schema = {
                'title': self._get_serializer_title(),
                'type': 'object',
                'required': self._required_fields(),
                'properties': self._get_all_field_properties(),
            }
        schema = self._add_dependencies(schema)
        return schema


class UiSchemaProcessor(ProcessingMixin):
    def _field_order(self) -> List[str]:
        if self._is_list_serializer(self.serializer):
            return list(self.serializer.child.Meta.fields)
        return list(self.serializer.Meta.fields)

    def _get_style_dict(self, field) -> Dict[str, Any]:
        style_dict = {}
        for k, v in field.style.items():
            if not k.startswith("schema:") and k not in STYLE_KEYS_TO_IGNORE:
                style_dict[k] = v
        return style_dict

    def _set_validation_properties(
        self, field: SerializerType, result: Dict[str, Any]
    ) -> Dict[str, Any]:
        excluded_validators = tuple(
            EXCLUDED_VALIDATOR_CLASSES + list(VALIDATION_MAP.keys())
        )
        custom_validators = [
            v for v in field.validators if not isinstance(v, excluded_validators)
        ]
        if custom_validators:
            result['ui:custom-validators'] = [
                {
                    'code': getattr(v, 'code', v.__class__.__name__.lower()),
                    'message': getattr(v, 'message'),
                }
                for v in custom_validators
            ]

        return result

    def _get_ui_field_properties(
        self, field: SerializerType, name: str
    ) -> Dict[str, Any]:
        data_index = self._generate_data_index(name)
        result = {}
        if self._is_field_serializer(field):
            return UiSchemaProcessor(
                field, self.renderer_context, prefix=data_index
            ).get_ui_schema()
        elif isinstance(field, serializers.ListField):
            child = field.child
            widget = self._get_type_map_value(field=child).get('widget')
            if not widget and isinstance(child, serializers.ChoiceField):
                widget = 'checkbox'
        else:
            widget = self._get_type_map_value(field=field).get('widget')
        help_text = field.help_text
        if widget:
            result['ui:widget'] = widget
        if help_text:
            result['ui:help'] = help_text
        result.update(self._get_style_dict(field))
        result = self._set_validation_properties(field, result)
        return result

    def _get_all_ui_properties(self) -> Dict[str, Any]:
        result = {}
        for name, field in self.fields:
            override = field.style.get(UI_SCHEMA_OVERRIDE_KEY)
            result[name] = override or self._get_ui_field_properties(field, name)
        return result

    def get_ui_schema(self) -> Dict[str, Any]:
        if self._is_list_serializer(self.serializer):
            return {
                'items': {
                    **{'ui:order': self._field_order()},
                    **self._get_style_dict(self.serializer),
                    **self._get_all_ui_properties(),
                }
            }
        else:
            return {
                **{'ui:order': self._field_order()},
                **self._get_style_dict(self.serializer),
                **self._get_all_ui_properties(),
            }


class ColumnProcessor(ProcessingMixin):
    def _get_column_properties(
        self, field: SerializerType, name: str
    ) -> Dict[str, str]:
        data_index = self._generate_data_index(name)
        result = {
            'title': self._get_title(field, data_index),
            'dataIndex': data_index,
            'key': name,
        }
        sort_order = field.style.get('schema:sort')
        if sort_order:
            if sort_order not in ['ascend', 'descend']:
                raise ValueError(
                    f"The {data_index} field 'style['schema:sort']' "
                    f"value must be either 'ascend' or 'descend'"
                )
            result['defaultSortOrder'] = sort_order
        return result

    def get_schema(self) -> List[Dict[str, str]]:
        result = []
        for name, field in self.fields:
            data_index = self._generate_data_index(name)
            if self._is_field_serializer(field):
                # TODO: How to list nested list serializers?
                if self._is_list_serializer(field):
                    continue
                result.extend(
                    ColumnProcessor(
                        field, self.renderer_context, prefix=data_index
                    ).get_schema()
                )
            else:
                override = field.style.get(COLUMN_PROCESSOR_OVERRIDE_KEY)
                result.append(override or self._get_column_properties(field, name))
        return result


class SerializerEncoder(DjangoJSONEncoder):
    LIST_ACTION = 'list'

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            skipkeys=kwargs.pop('skipkeys', False),
            ensure_ascii=kwargs.pop('ensure_ascii', True),
            check_circular=kwargs.pop('check_circular', True),
            allow_nan=kwargs.pop('allow_nan', True),
            indent=kwargs.pop('indent', None),
            separators=kwargs.pop('separators', None),
            default=kwargs.pop('default', None),
            sort_keys=kwargs.pop('sort_keys', False),
        )
        self.renderer_context = kwargs.pop('renderer_context', {})
        self.extra_kwargs = kwargs

    def _get_view_action(self) -> str:
        return self.renderer_context.get('view', {}).__dict__.get('action', '')

    def default(self, obj: Any) -> Union[Dict, List]:
        if isinstance(obj, serializers.Serializer):
            if self._get_view_action() == self.LIST_ACTION:
                return ColumnProcessor(obj, self.renderer_context).get_schema()
            else:
                return {
                    'schema': SchemaProcessor(obj, self.renderer_context).get_schema(),
                    'uiSchema': UiSchemaProcessor(
                        obj, self.renderer_context
                    ).get_ui_schema(),
                }
        return super().default(obj)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drf_react_template']

package_data = \
{'': ['*']}

install_requires = \
['djangorestframework>=3.12.0,<4.0.0']

setup_kwargs = {
    'name': 'drf-react-template-framework',
    'version': '0.0.16',
    'description': 'Django REST Framework plugin that creates form schemas for react-jsonschema-form',
    'long_description': '# Django Rest Framework React Template Framework\nDjango Rest Framework React Template Framework is a Python package that plugs into [Django Rest Framework](https://www.django-rest-framework.org/) to generate json schemas for [react-jsonschema-form](https://github.com/rjsf-team/react-jsonschema-form). The concept is similar to the DRF [HTMLFormRenderer](https://www.django-rest-framework.org/api-guide/renderers/#htmlformrenderer), but specifically designed for [React](https://reactjs.org/) frontends.\n\n## Installation\n```\npip install drf-react-template-framework\n```\n\n## Quick Start\n\nSee `/example` for a complete simple example.\n\n\n`serializers.py`:\n```python\nfrom rest_framework import serializers\n\nclass ChoiceSerializer(serializers.ModelSerializer):\n    choice_text = serializers.CharField()\n    votes = serializers.IntegerField()\n\n    class Meta:\n        fields = (\'choice_text\', \'votes\')\n```\n`viewsets.py`:\n```python\nfrom rest_framework.mixins import RetrieveModelMixin\nfrom drf_react_template.mixins import FormSchemaViewSetMixin\n\nclass ChoiceViewSet(\n    RetrieveModelMixin,\n    FormSchemaViewSetMixin,\n):\n    queryset = models.Choice.objects.all()\n    serializer_class = serializers.ChoiceSerializer\n```\nNow when the retrieve endpoint is called, `FormSchemaViewSetMixin` will update the response\nto include a `serializer` object in the json, while the original payload is sorted in `formData`.\n\nInside the `serializer` object, two fields `schema` and `uiSchema` contain the objects which\ncan be used inside [`react-jsonschema-form`](https://react-jsonschema-form.readthedocs.io/en/latest/#usage).\n\n## Documentation\n\nWhile the above quick start is useful for super simple cases,\nthe framework has a large number of additional parameters which can be used to customize behaviour.\n\n### Viewset\n\nAll viewsets must inherit from `drf_react_template.mixins.FormSchemaViewSetMixin`.\nThis class inherits from the\nDRF [`GenericViewSet`](https://www.django-rest-framework.org/api-guide/viewsets/#genericviewset),\nwhile additionally providing custom `finialize response functionality` and the additional endpoint:\n```\nGET */create/\n>> {\'serializer\': ..., \'formData\': {}}\n```\nWhich can be used to generate an empty form for create.\n\nThe only other specific customization that can be applied in the viewset is different\nserializers for different endpoints. For example, `update` actions often show a subset of fields;\nas such it is possible to override `get_serializer_class` to return the specific form required.\n\nFor example:\n```python\n# serializers.py\nclass ChoiceSerializer(serializers.ModelSerializer):\n    choice_text = serializers.CharField()\n    votes = serializers.IntegerField()\n\n    class Meta:\n        fields = (\'choice_text\', \'votes\')\n\nclass ChoiceUpdateSerializer(serializers.ModelSerializer):\n    choice_text = serializers.CharField()\n\n    class Meta:\n        fields = (\'choice_text\',)\n\n# viewsets.py\ndef get_serializer_class(self):\n    if self.action == \'update\':\n        return ChoiceUpdateSerializer\n    return ChoiceSerializer\n```\n\nSince this having a separate `list` serializer is so common, the above can be avoided by using\nthe `serializer_list_class` class attribute provided by `FormSchemaViewSetMixin`.\n\n### Serializer\n\nThe majority of the customization will occur inside serializer classes;\na real world example will often require custom `create`, `update`, `to_representation`, etc methods.\n\nAt the class level the serializer defines the form name:\n```\nChoiceSerializer -> Choice\nQuestionAndChoiceSerializer -> Question And Choice\nChoice -> Choice\n```\nIf a different title is required (or no title for nested forms), then the `__init__` method can be updated:\n```python\ndef __init__(self, *args, **kwargs):\n    kwargs[\'label\'] = \'New Form Title\'\n    super().__init__(*args, **kwargs)\n```\n\n#### Serializer Field Attributes\n\nThe following is a list of parameters that can be added to individual fields which modifies\n`react-jsonschema-form` functionality on the front-end.\n\n##### Label\nUpdates the form input title text. Can also be used to provide translations.\n```python\nchoice_text = serializers.CharField(label=\'What is the choice label?\')\n```\nThe label can also be disabled completely via `style`:\n```python\nchoice_text = serializers.CharField(style={\'ui:options\': {\'label\': False}})\n```\nNote: This does not work for `list` actions.\n\n##### Read Only\nSetting `read_only=True` forces the encoder to skip it when building the form,\nbut still means the value gets sent in the `formData`:\n```python\nchoice_text = serializers.CharField(read_only=True)\n```\nThis is useful for custom widgets which rely on data not related to the form (e.g. `id`).\n\n##### Required\nSetting `required=False` removes frontend validation of the field:\n```python\nchoice_text = serializers.CharField(required=False)\n```\nNote: This has no effect on `list` actions.\n\n##### Allow Null\nSetting `allow_null=True` allows `null` values to be sent back in `formData`:\n```python\nchoice_text = serializers.CharField(allow_null=True)\n```\nNote: This has no effect on `list` actions.\n\n##### Default\nBy setting a default value, a field is automatically set to `required=False`.\nThe `formData` object will also contain the `default` value if it is not provided one.\n```python\nchoice_text = serializers.CharField(default=\'example text\')\n```\nNote: This has no effect on `list` actions.\n\n##### Validation Mapping\nBuilt-in validators are automatically added to the schema. For instance:\n```python\nchar_text = serializers.CharField(min_length=5, max_length=10)\n```\n\nWill result in the following schema portion:\n\n```json\n{\n    "type": "string",\n    "title": "Char text",\n    "maxLength": 10,\n    "minLength": 5\n}\n```\n\nThe following core validators have mapping to json schema (https://ajv.js.org/json-schema.html):\n\n- MaxLengthValidator: maxLength\n- MinLengthValidator: minLength\n- MaxValueValidator: maximum\n- MinValueValidator: minimum\n- RegexValidator: pattern\n\nCustom validators can not be mapped directly to Ajv schema properies so they are rather provided in the json ui schema as hints.\n\n```python\nclass MinSizeImageValidator:\n    message = _(\'Image is too small, must be 1KB minimum.\')\n    code = \'image_min_1KB\'\n\n    def __call__(self, value):\n        min_size = 1024  # 1KB\n        if value.size < min_size:\n            raise serializers.ValidationError(self.message, code=self.code)\n\nimage_field = serializers.ImageField(\n    required=True, validators=[MinSizeImageValidator]\n)\n```\n\nWill result in the following ui schema portion:\n\n```json\n{\n    "ui:widget": "file",\n    "ui:custom-validators": [\n        {"code": "image_min_1KB", "message": "Image is too small, must be 1KB minimum."}\n    ]\n}\n```\n\n##### Style\nThe DRF `style` parameter is a `dict` and is therefore used for a number of different parameters.\nThere are a [number of options](https://react-jsonschema-form.readthedocs.io/en/latest/api-reference/uiSchema/)\nthat `react-jsonschema-form` provides, many of which should work out-of-the-box,\nalthough not all options will have been tested.\n\nSince style params are applied **last**, they can overwrite other keys. Additionally, any key not\nstarting with `schema:` will be sent in the `json` payload, so this can be a good place to send\nadditional attributes for custom widgets.\n\nThe following are a list of valid (tested) keys and their uses.\n\n**Note**: This can also be done at the serializer level:\n```python\ndef __init__(self, *args, **kwargs):\n    kwargs[\'style\'] = {\'ui:template\': \'Card\'}\n    super().__init__(*args, **kwargs)\n```\n\n###### Dependencies\n\n`react-jsonschema-form` allows for the use of [dependencies](https://react-jsonschema-form.readthedocs.io/en/latest/usage/dependencies/#dependencies)\nbetween multiple fields. This library has analogues for the four types (bidirectional, unidirectional, conditional, and\ndynamic).\n```python\n# Unidirectional\nchoice_text = serializers.CharField(\n    style={\'schema:dependencies:simple\': [\'votes\']}\n)\n\n# Bidirectional - note the different value styles, either is fine.\nchoice_text = serializers.CharField(\n    style={\'schema:dependencies:simple\': [\'votes\']}\n)\nvotes = serializers.IntegerField(\n    default=0, style={\'schema:dependencies:simple\': \'choice_text\'}\n)\n\n# Conditional\nchoice_text = serializers.CharField(\n    style={\'schema:dependencies:conditional\': [\'votes\']}\n)\n\n# Dynamic - must be an enum field, and only one choice per dependency,\n# multiple choices can have the same dependencies (side-stepping this issue).\nchoice_text = serializers.ChoiceField(\n    choices=((\'yes\', \'Yes\'), (\'no\', \'No\')),\n    style={\'schema:dependencies:dynamic\': {\'yes\': [\'votes\'], \'no\': None}}\n)\n```\n\n###### Field Overrides\n\nSince this framework doesn\'t cover 100% of the features of the `react-jsonschema-form`, it is possible to provide\n`dict` objects which override the individual field properties:\n```python\nchoice_text = serializers.CharField(\n    style={\n        \'schema:override\': {\'type\': \'string\', \'title\': \'Overridden title\'},\n        \'uiSchema:override\': {\'ui:widget\': \'updown\'},\n        \'column:override\': {\n            \'title\': \'Overridden title\',\n            \'dataIndex\': \'question_text\',\n            \'key\': \'question_text\',\n        },\n        \'schema:dependencies:override\': [\'votes\']\n    }\n)\n```\n**Note**: There is no validation around these overrides, so it is left up to the developer to ensure the resulting\nschema is valid. For example, `\'schema:dependencies:override\'` will not remove fields from the main `\'properties\'` or\n`\'required\'` objects. (This can be done by omitting the dependant field from the serializer or using `read_only=True`).\n\n###### List Sort Order\nSends the `defaultSortOrder` key with the `list` action serializer:\n```python\nchoice_text = serializers.CharField(, style={\'schema:sort\': \'ascend\'})\n```\n\n###### Widget, Type, and Enum\nWhile the framework tries to provide sensible defaults for DRF fields,\nsometimes custom frontend widgets need to provide custom behaviour.\n```python\nchoice_text = serializers.CharField(style={\n    \'ui:widget\': \'textarea\',\n    \'schema:type\': \'string\',\n    \'schema:enum\': \'choices\'\n})\n```\nMore information can be found on\n[`types`](https://react-jsonschema-form.readthedocs.io/en/latest/usage/single/#field-types)\nand [`enum`](https://react-jsonschema-form.readthedocs.io/en/latest/usage/single/#enumerated-values).\n\n`enum` can also be set to the `string` `choices` (as in the example),\nto try and use the `field.choices` attribute.\n\n###### Placeholder\nThe HTML placeholder text can be updated in the following way:\n```python\nchoice_text = serializers.CharField(style={\'ui:placeholder\': \'This a placeholder value\'})\n```\n\n###### Text Area Widget Rows\nWhen the `ui:widget` is set to `textarea` (see above), the `rows` can be set via:\n```python\nchoice_text = serializers.CharField(style={\n    \'ui:widget\': \'textarea\',\n    \'ui:options\': {\'rows\': 8},\n})\n```\n\n###### Disabled\nAn input field can be `disabled` in the following way:\n```python\nchoice_text = serializers.CharField(style={\'ui:disabled\': \'true\'})\n```\n\n### Settings\n\n##### DRF_REACT_TEMPLATE_TYPE_MAP\nAllows for custom fields to be added to `ProcessingMixin.TYPE_MAP` in this way:\n```python\nDRF_REACT_TEMPLATE_TYPE_MAP = {\n    \'ImageField\': {\'type\': \'image\', \'widget\': \'file\'},\n}\n```\nThe default supported field types are:\n\n```python\n{\n    \'CharField\': {\'type\': \'string\'},\n    \'IntegerField\': {\'type\': \'integer\', \'widget\': \'updown\'},\n    \'FloatField\': {\'type\': \'number\', \'widget\': \'updown\'},\n    \'DecimalField\': {\'type\': \'number\', \'widget\': \'updown\'},\n    \'BooleanField\': {\'type\': \'boolean\'},\n    \'DateTimeField\': {\'type\': \'string\', \'widget\': \'date-time\'},\n    \'DateField\': {\'type\': \'string\', \'widget\': \'date\'},\n    \'URLField\': {\'type\': \'string\', \'widget\': \'uri\'},\n    \'ChoiceField\': {\'type\': \'string\', \'enum\': \'choices\'},\n    \'EmailField\': {\'type\': \'string\', \'widget\': \'email\'},\n    \'ListField\': {\'type\': \'array\'},\n}\n```\n\n## Development\n\nThis Repo uses [Poetry](https://python-poetry.org/docs/),\nand setup for development is as simple as follows:\n```bash\npoetry init\npre-commit install\n```\n',
    'author': 'stuart.bradley',
    'author_email': 'stuy.bradley@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yoyowallet/drf-react-template-framework',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from rest_framework import status
from rest_framework.decorators import action
from rest_framework.mixins import Response
from rest_framework.viewsets import GenericViewSet

from drf_react_template.renderers import JSONSerializerRenderer


class FormSchemaViewSetMixin(GenericViewSet):
    renderer_classes = (JSONSerializerRenderer,)
    serializer_list_class = None

    def get_serializer_class(self):
        if self.action == 'list' and self.serializer_list_class:
            return self.serializer_list_class
        return self.serializer_class

    def finalize_response(self, request, response, *args, **kwargs):
        response = super(FormSchemaViewSetMixin, self).finalize_response(
            request, response, args, kwargs
        )
        if response.status_code in (status.HTTP_200_OK, status.HTTP_201_CREATED):
            response.data = {
                'serializer': self.get_serializer(),
                'formData': response.data,
            }
        return response

    @action(detail=False, methods=('get',), url_path='create')
    def create_form(self, request, *args, **kwargs):
        return Response({})
